#include <stdio.h>
#include <fcntl.h>
#include <CoreServices/CoreServices.h>
#include <CoreAudio/CoreAudio.h>
#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/AudioToolbox.h>
#include <unistd.h>
#include <pthread.h>
#include "openspc.h"

#define SOFTBUFSIZE 65536

OSStatus MyFileRenderProc(void *inRefCon, AudioUnitRenderActionFlags    *inActionFlags,
	const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
	UInt32 inNumFrames, AudioBufferList *ioData);
OSStatus MyACComplexInputProc   (AudioConverterRef                        inAudioConverter,
	UInt32                               *ioNumberDataPackets,
	AudioBufferList                                   *ioData,
	AudioStreamPacketDescription  **outDataPacketDescription,
	void*                                          inUserData);
void convertToAIFF(char *file);
int readID666Text(char *file);
int readID666Bin(char *file);
void *receiveKeyInput(void* param);

typedef struct {
	char FORM[4];
	int formChunkSize;
	char AIFF[4];
	char COMM[4];
	int commChunkSize;
	short numChannels;
	unsigned int numSampleFrames;
	short sampleSize;
	int sampleRate1;
	int sampleRate2;
	short sampleRate3;
	char SSND[4];
	int ssndChunkSize;
	unsigned int offset;
	unsigned int blockSize;
} AIFFHeader;

static AudioUnit outputUnit;
static AudioUnit reverbUnit;
static AudioConverterRef converter;
static char *buffer1;
static char *buffer2;
static char *buffer1Top;
void *gSourceBuffer;
static int bufferIsFull;
static int buffer1Rest;
static int buffer2Rest;
static int totalSize;
static int playedSize;
static float totalTime;
static int currentTime;
static float fadeOut;
static int lastBuffer;
static int playDone;
static int first;
static int convertMode;
static FILE *fpw;
static openspc ospc;
static float defaultPlayTime;
static float defaultFadeOut;
static int force;
static pthread_t tid;
static unsigned char channelMask;
static int interpolation;
static int detectSilence;
static int silence;
static int silenceLimit;

int main(int argc, char *argv[])
{
	OSStatus	err ;
	AudioStreamBasicDescription inFormat;
	AudioStreamBasicDescription outFormat;
	AURenderCallbackStruct  renderCallback;
	Boolean outWritable;
	UInt32		count ;
	int i=0;
	int reverb = 0;
	int reverbType = 0;
	float reverbStrength = 100.0;
	FILE *fp;
	char opt, buf[4];
	extern char *optarg ;
	extern int optind;
	defaultPlayTime = 120;
	defaultFadeOut = 10;
	force = 0;
	channelMask = 255;
	interpolation = 0;
	silence = 0;
	detectSilence = 0;
	silenceLimit = 384000;
	
	while((opt = getopt(argc,argv,"o:t:f:Fm:r:s:i:c:")) != -1) {
		switch(opt) {
		  case 'o' :
			convertMode = 1;
			if(!(fpw=fopen(optarg,"wb"))) {
				fprintf(stderr,"can't open file.\n");
				return 0;
			}
			i++;
			break;
		  case 't' :
			defaultPlayTime = atof(optarg);
			if(defaultPlayTime <= 0) defaultPlayTime = 120;
			break;
		  case 'f' :
			defaultFadeOut = atof(optarg);
			if(defaultFadeOut < 0) defaultFadeOut = 10;
			break;
		  case 'F' :
			force = 1;
			break;
		  case 'm' :
			channelMask = atoi(optarg);
			break;
		  case 'r' :
			reverbType = atoi(optarg);
			reverb = 1;
			break;
		  case 's' :
			reverbStrength = atof(optarg);
			if(reverbStrength < 0 || reverbStrength > 100) reverbStrength = 100;
			break;
		  case 'i' :
			interpolation = atoi(optarg);
			if(interpolation != 0 && interpolation != 1) interpolation = 0;
			break;
		  case 'c' :
			detectSilence = 1;
			silenceLimit = atoi(optarg)*32000*4;
			if(silenceLimit <= 0) silenceLimit = 384000;
			break;
		}
		i++;
	}
	
	
	if(argc <= i+1) {
		printf("usage: %s <SPC file>\n",argv[0]);
		return 0;
	}
	
	if(!(fp=fopen(argv[optind],"rb"))) {
		fprintf(stderr,"can't open file.\n");
		return 0;
	}
	
	if(fread(buf,1,4,fp)!=4) {
		fprintf(stderr,"can't read file.\n");
		fclose(fp);
		return 0;
	}
	
	fclose(fp);
	
	if(memcmp(buf,"SNES",4))
	{
		fprintf(stderr,"file is not SPC format.\n");
		return 0;
	}
	
	
	if(convertMode) {
		convertToAIFF(argv[optind]);
		return 0;
	}
	
	ComponentDescription desc;
	Component comp;
	AudioUnitConnection	connection;
	connection.sourceOutputNumber = 0;
	connection.destInputNumber    = 0;
	
	/* Open default output nuit */
	desc.componentType = kAudioUnitType_Output; 
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	comp = FindNextComponent(NULL, &desc);
	if (comp == NULL) {
		fprintf(stderr, "can't find default output unit\n");
		return 0;
	}
	
	err = OpenAComponent(comp, &outputUnit);
	if (err)  {
		fprintf(stderr, "can't open default output unit\n");
		return 0;
	}
	
	/* open matrixreverb unit */
	desc.componentType = kAudioUnitType_Effect; 
	desc.componentSubType = kAudioUnitSubType_MatrixReverb;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	comp = FindNextComponent(NULL, &desc);
	if (comp == NULL) {
		fprintf(stderr, "can't find default output unit\n");
		return 0;
	}
	
	err = OpenAComponent(comp, &reverbUnit);
	if (err)  {
		fprintf(stderr, "can't open default output unit\n");
		return 0;
	}
	
	err = AudioUnitInitialize(outputUnit);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitInitialize failed.\n");
		return 0;
	}
	err = AudioUnitInitialize(reverbUnit);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitInitialize failed.\n");
		return 0;
	}
	
	AudioUnit *target;
	if(reverb) target = &reverbUnit;
	else target = &outputUnit;
	
	if(reverb) {
		connection.sourceAudioUnit = reverbUnit;
		err = AudioUnitSetProperty(outputUnit, kAudioUnitProperty_MakeConnection, kAudioUnitScope_Input, 0, &connection, sizeof(connection));
		if(err != noErr) {
			fprintf(stderr, "AudioUnitSetProperty failed.\n");
			return 0;
		}
		err = AudioUnitSetProperty (reverbUnit, kAudioUnitProperty_ReverbRoomType, kAudioUnitScope_Global, 0, &reverbType, sizeof(reverbType));
		if(err != noErr) {
			fprintf(stderr, "AudioUnitSetProperty failed.\n");
		return 0;
		}
		AudioUnitSetParameter(reverbUnit, kReverbParam_DryWetMix, kAudioUnitScope_Global, 0, reverbStrength, 0);
	}
		
	err = AudioUnitGetPropertyInfo(*target, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &count, &outWritable);
	err = AudioUnitGetProperty (*target, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &outFormat, &count);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitGetProperty failed.\n");
		return 0;
	}
	AudioUnitSetProperty (*target, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &outFormat, count);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitSetProperty failed.\n");
		return 0;
	}
	
	
	inFormat.mSampleRate = 32000.0;
	inFormat.mBytesPerPacket = 4;
	inFormat.mFramesPerPacket = 1;
	inFormat.mBytesPerFrame = 4;
	inFormat.mChannelsPerFrame = 2;
	inFormat.mBitsPerChannel = 16;
	inFormat.mFormatID = kAudioFormatLinearPCM;
	inFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked | kLinearPCMFormatFlagIsBigEndian;
	
	err = AudioConverterNew(&inFormat, &outFormat, &converter);
	if(err != noErr) {
		fprintf(stderr, "AudioConverterNew failed.\n");
		return 0;
	}
	
	memset(&renderCallback, 0, sizeof(AURenderCallbackStruct));
	renderCallback.inputProc = MyFileRenderProc;
	renderCallback.inputProcRefCon =0;
	err = AudioUnitSetProperty (*target, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input, 0, &renderCallback, sizeof(AURenderCallbackStruct));
	if(err != noErr) {
		fprintf(stderr, "AudioUnitSetProperty failed.\n");
		return 0;
	}
	
	int fd=open(argv[optind],O_RDONLY);
	int size=lseek(fd,0,SEEK_END);
	lseek(fd,0,SEEK_SET);
	void *ptr=malloc(size);
	read(fd,ptr,size);
	close(fd);
	
	if((fd=ospc.init(ptr,size)))
	{
		fprintf(stderr,"\nospc.Init returned %d!\n",fd);
		return 0;
	}
	
	free(ptr);
	
	fp=fopen(argv[optind], "rb");
	fseek(fp,0x23,SEEK_SET);
	char c;
	fread(&c,1,1,fp);
	if(c != 0x1a) force = 1;
	fclose(fp);
	
	if(!force) {
		fp=fopen(argv[optind], "rb");
		fseek(fp,0xd2,SEEK_SET);
		fread(&c,1,1,fp);
		fclose(fp);
		if(c >= 0x30) {
			int ret = readID666Text(argv[optind]);
			if(ret) readID666Bin(argv[optind]);
		}
		else {
			int ret = readID666Bin(argv[optind]);
			if(ret) readID666Text(argv[optind]);
		}
	}
	else {
		totalTime = defaultPlayTime;
		fadeOut = defaultFadeOut;
		totalSize = (int)(totalTime * 32000 * 4);
	}
	
	if(fadeOut > totalTime) fadeOut = totalTime;
	
	//printf("%f, %f\n",totalTime, fadeOut);
	
	if((buffer1 = (char *)malloc(SOFTBUFSIZE)) == NULL) {
		fprintf(stderr,"malloc error\n");
		return 0;
	}
	if((buffer2 = (char *)malloc(SOFTBUFSIZE)) == NULL) {
		fprintf(stderr,"malloc error\n");
		return 0;
	}
	
	bufferIsFull = 0;
	buffer1Rest = SOFTBUFSIZE;
	buffer2Rest = 0;
	buffer1Top = buffer1;
	lastBuffer = 0;
	playDone = 0;
	first = 1;
	playedSize = 0;
	currentTime = -1;
	ospc.setDefaultMuteChannel(channelMask);
	ospc.setInterpolationMethod(interpolation);
	
	//AudioUnitSetParameter(outputUnit, kAudioUnitParameterUnit_LinearGain, kAudioUnitScope_Global, 0, 1.4, 0);
	
	pthread_create(&tid,NULL,receiveKeyInput,NULL);
	pthread_detach(tid);
	
	AudioOutputUnitStart(outputUnit);
	/*
	while(1) {
		while(bufferIsFull && !lastBuffer) usleep(10000);
		if(lastBuffer) break;
		size = ospc.run(-1,(short *)buffer1,SOFTBUFSIZE);
		//if(size < SOFTBUFSIZE) break;
		bufferIsFull = 1;
		if(first) {
			first = 0;
			AudioOutputUnitStart(outputUnit);
		}
	}
	*/
	/*
	memset(buffer1+size, 0, SOFTBUFSIZE-size);
	bufferIsFull = 1;
	lastBuffer = 1;
	*/
	
	
	while(!playDone) {
		usleep(10000);
	}
	
	AudioConverterDispose(converter);
	AudioOutputUnitStop(outputUnit);
	AudioUnitUninitialize (outputUnit);
	CloseComponent(outputUnit);
	return 0;
}

void convertToAIFF(char *file)
{
	AIFFHeader header;
	playedSize = 0;
	currentTime = -1;
	int ret, i;
	
	int fd=open(file,O_RDONLY);
	int size=lseek(fd,0,SEEK_END);
	lseek(fd,0,SEEK_SET);
	void *ptr=malloc(size);
	read(fd,ptr,size);
	close(fd);
	
	if((fd=ospc.init(ptr,size)))
	{
		fprintf(stderr,"\nospc.Init returned %d!\n",fd);
		return;
	}
	
	free(ptr);
	
	if(!force) {
		FILE *fp=fopen(file, "rb");
		char c;
		fseek(fp,0xd2,SEEK_SET);
		fread(&c,1,1,fp);
		fclose(fp);
		if(c >= 0x30) {
			ret = readID666Text(file);
			if(ret) readID666Bin(file);
		}
		else {
			ret = readID666Bin(file);
			if(ret) readID666Text(file);
		}
	}
	else {
		totalTime = defaultPlayTime;
		fadeOut = defaultFadeOut;
		totalSize = (int)(totalTime * 32000 * 4);
	}
	
	if(fadeOut > totalTime) fadeOut = totalTime;
	ospc.setDefaultMuteChannel(channelMask);
	ospc.setInterpolationMethod(interpolation);
	
	memcpy(header.FORM, "FORM", 4);
	fwrite(&header.FORM, 1, 4, fpw);
	header.formChunkSize = totalSize + 0x2e;
	fwrite(&header.formChunkSize, 4, 1, fpw);
	memcpy(header.AIFF, "AIFF", 4);
	fwrite(&header.AIFF, 1, 4, fpw);
	memcpy(header.COMM, "COMM", 4);
	fwrite(&header.COMM, 1, 4, fpw);
	header.commChunkSize = 0x12;
	fwrite(&header.commChunkSize, 4, 1, fpw);
	header.numChannels = 2;
	fwrite(&header.numChannels, 2, 1, fpw);
	header.numSampleFrames = totalSize/4;
	fwrite(&header.numSampleFrames, 4, 1, fpw);
	header.sampleSize = 0x10;
	fwrite(&header.sampleSize, 2, 1, fpw);
	header.sampleRate1 = 0x400dfa00;
	fwrite(&header.sampleRate1, 4, 1, fpw);
	header.sampleRate2 = 0;
	fwrite(&header.sampleRate2, 4, 1, fpw);
	header.sampleRate3 = 0;
	fwrite(&header.sampleRate3, 2, 1, fpw);
	memcpy(header.SSND, "SSND", 4);
	fwrite(&header.SSND, 1, 4, fpw);
	header.ssndChunkSize = totalSize + 8;
	fwrite(&header.ssndChunkSize, 4, 1, fpw);
	header.offset = 0;
	fwrite(&header.offset, 4, 1, fpw);
	header.blockSize = 0;
	fwrite(&header.blockSize, 4, 1, fpw);
	
	char *buffer = (char *)malloc(32768);
	short *shrtPtr = (short *)buffer;
	float sec;
	
	while(1) {
		sec = ((float)(playedSize+512))/32000/4;
		if(sec >= totalTime - fadeOut && fadeOut) {
			ret = ospc.run(-1,(short *)buffer,512);
			float scale = 1.0f - (sec - (totalTime - fadeOut))*(1.0f/fadeOut);
			if(scale < 0) scale = 0;
			if(scale > 1) scale = 1;
			for(i=0;i<256;i++) {
				shrtPtr[i] = (short)(shrtPtr[i] * scale);
			}
		}
		else {
			ret = ospc.run(-1,(short *)buffer,32768);
		}
		playedSize = playedSize + ret;
		if(playedSize >= totalSize) break;
		fwrite(buffer, 1, ret, fpw);
		
		int percent = (int)(100.0*playedSize/totalSize);
		if(percent > 100) percent = 100;
		if(percent != currentTime) {
			printf("\r|");
			for(i=0;i<20;i++) {
				if(percent/5 > i)
					printf("=");
				else if(percent/5 == i)
					printf(">");
				else printf("-");
			}
			printf("| %d%%",percent);
			fflush(stdout);
			currentTime = percent;
		}
	}
	
	fwrite(buffer, 1, ret - (playedSize - totalSize), fpw);
	
	header.formChunkSize = totalSize + 0x2e;
	header.numSampleFrames = totalSize/4;
	header.ssndChunkSize = totalSize + 8;
	fseek(fpw, 4, SEEK_SET);
	fwrite(&header.formChunkSize, 4, 1, fpw);
	fseek(fpw, 0xe, SEEK_CUR);
	fwrite(&header.numSampleFrames, 4, 1, fpw);
	fseek(fpw, 0x10, SEEK_CUR);
	fwrite(&header.ssndChunkSize, 4, 1, fpw);
	
	fclose(fpw);
	
	free(buffer);
}


OSStatus MyFileRenderProc(void *inRefCon, AudioUnitRenderActionFlags    *inActionFlags,
                        const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
                        UInt32 inNumFrames, AudioBufferList *ioData)
{
	OSStatus err= noErr;
	void *inInputDataProcUserData=NULL;
	AudioStreamPacketDescription* outPacketDescription =NULL;
	
	err = AudioConverterFillComplexBuffer(converter, MyACComplexInputProc ,inInputDataProcUserData , &inNumFrames, ioData, outPacketDescription);
	if(!playDone && err != noErr) {
		fprintf(stderr, "AudioConverterFillComplexBuffer failed.\n");
		return 0;
	}
	
	return err;
}

OSStatus MyACComplexInputProc   (AudioConverterRef                        inAudioConverter,
                                 UInt32                               *ioNumberDataPackets,
                                 AudioBufferList                                   *ioData,
                                 AudioStreamPacketDescription  **outDataPacketDescription,
                                 void*                                          inUserData)
{
	OSStatus    err = noErr;
	UInt32  bytesCopied = 0;
	int i, percent;
	float sec;
	
	ioData->mBuffers[0].mData = NULL;      
	//ioData->mBuffers[0].mDataByteSize = 0;
	
	if(lastBuffer) {
		playDone = 1;
		return noErr;
	}
	/*
	if(!buffer2Rest) {
		while(!bufferIsFull) usleep(10000);
		memcpy(buffer2, buffer1, SOFTBUFSIZE);
		buffer1Rest = SOFTBUFSIZE;
		buffer1Top = buffer1;
		bufferIsFull = 0;
		buffer2Rest = SOFTBUFSIZE;
	}
	*/
	
	if (gSourceBuffer != NULL) {
		free(gSourceBuffer);
		gSourceBuffer = NULL;
	}
	
	gSourceBuffer = (void *) calloc (1, *ioNumberDataPackets * 4);
	bytesCopied = ospc.run(-1,(short *)gSourceBuffer,*ioNumberDataPackets * 4);
	ioData->mBuffers[0].mData = gSourceBuffer;
	ioData->mBuffers[0].mDataByteSize = bytesCopied;
	if(detectSilence) {
		for(i=0;i<bytesCopied;i+=2) {
			if(*((unsigned short *)(gSourceBuffer)+i/2) != 0 && *((unsigned short *)(gSourceBuffer)+i/2) < 0xFFFE) break;
		}
		silence = silence + i;
		if(silence >= silenceLimit) lastBuffer = 1;
		else if(i < bytesCopied) silence = 0;
	}
	
	//printf("packet size: %d\n",*ioNumberDataPackets);
	/*
	if (*ioNumberDataPackets > buffer2Rest / 4) 
		*ioNumberDataPackets = buffer2Rest / 4;
	
	if (*ioNumberDataPackets)
	{
		if (gSourceBuffer != NULL) {
			free(gSourceBuffer);
			gSourceBuffer = NULL;
		}
		
		bytesCopied = *ioNumberDataPackets * 4;  
		gSourceBuffer = (void *) calloc (1, bytesCopied);             
		memcpy(gSourceBuffer, buffer2 + SOFTBUFSIZE - buffer2Rest, bytesCopied);
		
		buffer2Rest = buffer2Rest - bytesCopied;
		
		ioData->mBuffers[0].mData = gSourceBuffer;
		ioData->mBuffers[0].mDataByteSize = bytesCopied;
	}  
	else
	{
		ioData->mBuffers[0].mData = NULL;      
		ioData->mBuffers[0].mDataByteSize = 0;
		err = noErr;
	}
	*/
	playedSize = playedSize + bytesCopied;
	if(playedSize >= totalSize) {
		playedSize = totalSize;
		lastBuffer = 1;
	}
	sec = (float)playedSize / 32000 / 4;
	int secint = (int)sec;
	
	if(sec >= totalTime - fadeOut && fadeOut) {
		float scale = 1.0f - (sec - (totalTime - fadeOut))*(1.0f/fadeOut);
		if(scale < 0) scale = 0;
		if(scale > 1.0) scale = 1.0;
		AudioUnitSetParameter(outputUnit, kAudioUnitParameterUnit_LinearGain, kAudioUnitScope_Global, 0, scale, 0);
	}
	
	if(secint != currentTime) {
		percent = (int)(100.0*playedSize/totalSize);
		printf("\r|");
		for(i=0;i<20;i++) {
			if(percent/5 > i)
				printf("=");
			else if(percent/5 == i)
				printf(">");
			else printf("-");
		}
		printf("| %d:%02d/%d:%02d (%d%%)",secint/60, secint-60*(secint/60), (int)totalTime/60, (int)totalTime-60*((int)totalTime/60), percent);
		fflush(stdout);
		currentTime = secint;
	}
	
	//if(!buffer2Rest && lastBuffer && !bufferIsFull) playDone = 1;
	
	return err;
}

int readID666Text(char *file)
{
	FILE *fp;
	char buf[6];
	float playTime;
	int ret=0;
	
	fp=fopen(file, "rb");
	fseek(fp,0xa9,SEEK_SET);
	fread(buf,1,3,fp);
	buf[3] = '\0';
	if(buf[0]) {
		playTime = atoi(buf);
		if(playTime <= 0) ret = 1;
	}
	else {
		playTime = defaultPlayTime - defaultFadeOut;
	}
	fread(buf,1,5,fp);
	buf[5] = '\0';
	if(buf[0]) {
		fadeOut = atoi(buf) / 1000.0f;
		if(fadeOut < 0) ret = 1;
	}
	else {
		fadeOut = defaultFadeOut;
	}
	
	totalTime = playTime + fadeOut;
	totalSize = (int)(totalTime * 32000 * 4);
	fclose(fp);
	
	return ret;
}

int readID666Bin(char *file)
{
	FILE *fp;
	short tmp;
	int tmp2, ret=0;
	float playTime;
	
	fp=fopen(file, "rb");
	fseek(fp,0xa9,SEEK_SET);
	fread(&tmp,2,1,fp);
	if(tmp) {
		playTime = ((tmp >> 8) & 0xFF) | ((tmp << 8) & 0xFF00);
		if(playTime <= 0) ret = 1;
	}
	else {
		playTime = defaultPlayTime - defaultFadeOut;
	}
	fseek(fp,1,SEEK_CUR);
	fread(&tmp2,4,1,fp);
	if(tmp) {
		fadeOut = (((tmp2 >> 24) & 0xFF) | ((tmp2 >> 8) & 0xFF00) | ((tmp2 << 8) & 0xFF0000) | ((tmp2 << 24) & 0xFF000000)) / 1000.0f;
		if(fadeOut < 0) ret = 1;
	}
	else {
		fadeOut = defaultFadeOut;
	}
	
	totalTime = playTime + fadeOut;
	totalSize = (int)(totalTime * 32000 * 4);
	fclose(fp);
	
	return ret;
}

void *receiveKeyInput(void* param)
{
	int c;
	while(!playDone) {
		c = getchar_unlocked();
		if(c >= '0' && c <= '7') {
			ospc.toggleChannelMute(c-0x30);
		}
		//usleep(10000);
	}
	return NULL;
}